# -*- coding: utf-8 -*-

import signal
import os
# linux 默认就是epoll
from twisted.internet import reactor
import setproctitle

from netkit.box import Box
from connection.client_tcp_connection import ClientConnectionFactory as ClientTCPConnectionFactory
from connection.client_udp_connection import ClientConnectionFactory as ClientUDPConnectionFactory
from connection.worker_connection import WorkerConnectionFactory
from connection.admin_connection import AdminConnectionFactory
from connection.master_connection import MasterConnectionFactory
from task_dispatcher import TaskDispatcher
from stat_counter import StatCounter
from ..share import constants
from ..share.log import logger


class Proxy(object):
    """
    proxy相关
    """

    type = constants.PROC_TYPE_PROXY

    client_tcp_connection_factory_class = ClientTCPConnectionFactory
    client_udp_connection_factory_class = ClientUDPConnectionFactory
    worker_connection_factory_class = WorkerConnectionFactory
    admin_connection_factory_class = AdminConnectionFactory
    master_connection_factory_class = MasterConnectionFactory

    app = None

    # 任务调度器
    task_dispatcher = None
    # 统计
    stat_counter = None

    # master的连接，因为一定只有一个，所以就一个变量即可
    master_conn = None

    def __init__(self, app):
        """
        构造函数
        :return:
        """
        self.app = app

        self.task_dispatcher = TaskDispatcher(self, self._on_workers_reload_over)
        self.stat_counter = StatCounter(self.app.config['TASKS_TIME_BENCHMARK'])

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name(self.type))

        # 主进程
        self._handle_proc_signals()

        ipc_directory = self.app.config['IPC_ADDRESS_DIRECTORY']
        if not os.path.exists(ipc_directory):
            os.makedirs(ipc_directory)

        # 启动监听master
        master_address = os.path.join(ipc_directory, self.app.config['MASTER_ADDRESS'])
        if os.path.exists(master_address):
            os.remove(master_address)
        reactor.listenUNIX(master_address, self.master_connection_factory_class(self))

        # 启动监听worker
        for group_id in self.app.config['GROUP_CONFIG']:
            ipc_address = os.path.join(ipc_directory, self.app.config['WORKER_ADDRESS_TPL'] % group_id)

            # 防止之前异常导致遗留
            if os.path.exists(ipc_address):
                os.remove(ipc_address)

            # 给内部worker通信用的
            reactor.listenUNIX(ipc_address, self.worker_connection_factory_class(self, group_id))

        # 启动对外监听
        host_list = self.app.config['HOST']
        if isinstance(host_list, (str, unicode)):
            # 说明是字符串，要转化成数组
            host_list = (host_list,)

        if self.app.config['TCP']:
            for host in host_list:
                reactor.listenTCP(self.app.config['PORT'], self.client_tcp_connection_factory_class(self),
                                  backlog=self.app.config['PROXY_BACKLOG'], interface=host)

        if self.app.config['UDP']:
            for host in host_list:
                reactor.listenUDP(
                    self.app.config['PORT'], self.client_udp_connection_factory_class(self),
                    interface=host
                )

        # 启动admin服务
        admin_address = self.app.config['ADMIN_ADDRESS']

        if admin_address:
            if isinstance(admin_address, (list, tuple)):
                # 说明是网络协议
                reactor.listenTCP(admin_address[1], self.admin_connection_factory_class(self),
                                  interface=admin_address[0])
            elif isinstance(admin_address, str):
                # 说明是文件
                admin_address = os.path.join(ipc_directory, admin_address)
                # 防止之前异常导致遗留
                if os.path.exists(admin_address):
                    os.remove(admin_address)
                reactor.listenUNIX(admin_address, self.admin_connection_factory_class(self))
            else:
                logger.error('invalid admin address. proxy: %s, admin_address: %s', self, admin_address)

        try:
            reactor.run(installSignalHandlers=False)
        except KeyboardInterrupt:
            pass
        except:
            logger.error('exc occur. proxy: %s', self, exc_info=True)

    def _handle_proc_signals(self):
        def stop_handler(signum, frame):
            """
            在centos6下，callFromThread(stop)无效，因为处理不够及时
            """
            try:
                reactor.stop()
            except:
                pass

        def safe_reload_handler(signum, frame):
            """
            让所有子进程重新加载
            """
            self.task_dispatcher.start_reload()

        # 强制结束，抛出异常终止程序进行
        signal.signal(signal.SIGINT, stop_handler)
        signal.signal(signal.SIGQUIT, stop_handler)
        # 直接停止
        signal.signal(signal.SIGTERM, stop_handler)
        # 忽略，因为这个时候是在重启worker
        signal.signal(signal.SIGHUP, safe_reload_handler)

    def _on_workers_reload_over(self):
        """
        当workers reload之后的操作
        需要给master通知，让master替换掉workers
        :return:
        """
        if self.master_conn and self.master_conn.transport:
            box = Box(dict(
                cmd=constants.CMD_MASTER_REPLACE_WORKERS
            ))
            self.master_conn.transport.write(box.pack())

    def __repr__(self):
        return '<%s name: %s>' % (
            type(self).__name__, self.app.name
        )



# -*- coding: utf-8 -*-

from collections import defaultdict
from ..share import constants


class ReloadHelper(object):

    proxy = None
    status = None

    # 预备役workers，使用dict，可以保证判断的时候更准确
    workers_dict = None

    def __init__(self, proxy):
        self.status = constants.RELOAD_STATUS_STOPPED
        self.proxy = proxy
        self.workers_dict = defaultdict(set)

    def start(self):
        """
        启动
        :return:
        """
        if self.status != constants.RELOAD_STATUS_STOPPED:
            return False

        self.status = constants.RELOAD_STATUS_PREPARING
        return True

    def stop(self):
        """
        停止
        :return:
        """
        self.status = constants.RELOAD_STATUS_STOPPED
        self.workers_dict.clear()

    def add_worker(self, worker):
        """
        添加worker
        :param worker:
        :return:
        """
        self.workers_dict[worker.group_id].add(worker)

        if self.status == constants.RELOAD_STATUS_PREPARING and self._match_expect_workers():
            self.status = constants.RELOAD_STATUS_WORKERS_DONE

    def remove_worker(self, worker):
        """
        添加worker
        :param worker:
        :return:
        """
        if worker in self.workers_dict[worker.group_id]:
            self.workers_dict[worker.group_id].remove(worker)

            if self.status == constants.RELOAD_STATUS_WORKERS_DONE and not self._match_expect_workers():
                self.status = constants.RELOAD_STATUS_PREPARING

            return True
        else:
            return False

    def _match_expect_workers(self):
        """
        是否满足预期的workers
        :return:
        """

        for group_id, group_info in self.proxy.app.config['GROUP_CONFIG'].items():
            expect_count = group_info['count']

            if len(self.workers_dict[group_id]) != expect_count:
                # 只要找到一个没有满足的，就可以扔掉了
                return False
        else:
            return True

    @property
    def running(self):
        return self.status in (constants.RELOAD_STATUS_PREPARING, constants.RELOAD_STATUS_WORKERS_DONE)

    @property
    def workers_done(self):
        """
        workers是否已经准备好了
        :return:
        """
        return self.status == constants.RELOAD_STATUS_WORKERS_DONE



# -*- coding: utf-8 -*-
"""
    modify from flask.config
    ~~~~~~~~~~~~

    Implements the configuration related objects.

    :copyright: (c) 2011 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""

import sys
import imp
import os
import errno

PY2 = sys.version_info[0] == 2

if not PY2:
    string_types = (str, )
else:
    string_types = (str, unicode)


def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    # force the import name to automatically convert to strings
    # __import__ is not able to handle unicode strings in the fromlist
    # if the module is a package
    import_name = str(import_name).replace(':', '.')
    try:
        try:
            __import__(import_name)
        except ImportError:
            if '.' not in import_name:
                raise
        else:
            return sys.modules[import_name]

        module_name, obj_name = import_name.rsplit('.', 1)
        try:
            module = __import__(module_name, None, None, [obj_name])
        except ImportError:
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            module = import_string(module_name)

        try:
            return getattr(module, obj_name)
        except AttributeError as e:
            raise ImportError(e)

    except ImportError as e:
        if not silent:
            t, v, tb = sys.exc_info()
            raise t, v, tb


class ConfigAttribute(object):
    """Makes an attribute forward to the config"""

    def __init__(self, name, get_converter=None):
        self.__name__ = name
        self.get_converter = get_converter

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        rv = obj.config[self.__name__]
        if self.get_converter is not None:
            rv = self.get_converter(rv)
        return rv

    def __set__(self, obj, value):
        obj.config[self.__name__] = value


class Config(dict):
    """Works exactly like a dict but provides ways to fill it from files
    or special dictionaries.  There are two common patterns to populate the
    config.

    Either you can fill the config from a config file::

        app.config.from_pyfile('yourconfig.cfg')

    Or alternatively you can define the configuration options in the
    module that calls :meth:`from_object` or provide an import path to
    a module that should be loaded.  It is also possible to tell it to
    use the same module and with that provide the configuration values
    just before the call::

        DEBUG = True
        SECRET_KEY = 'development key'
        app.config.from_object(__name__)

    In both cases (loading from any Python file or loading from modules),
    only uppercase keys are added to the config.  This makes it possible to use
    lowercase values in the config file for temporary values that are not added
    to the config or to define the config keys in the same file that implements
    the application.

    Probably the most interesting way to load configurations is from an
    environment variable pointing to a file::

        app.config.from_envvar('YOURAPPLICATION_SETTINGS')

    In this case before launching the application you have to set this
    environment variable to the file you want to use.  On Linux and OS X
    use the export statement::

        export YOURAPPLICATION_SETTINGS='/path/to/config/file'

    On windows use `set` instead.

    :param root_path: path to which files are read relative from.  When the
                      config object is created by the application, this is
                      the application's :attr:`~flask.Flask.root_path`.
    :param defaults: an optional dictionary of default values
    """

    def __init__(self, root_path=None, defaults=None):
        dict.__init__(self, defaults or {})
        self.root_path = root_path or ''

    def from_envvar(self, variable_name, silent=False):
        """Loads a configuration from an environment variable pointing to
        a configuration file.  This is basically just a shortcut with nicer
        error messages for this line of code::

            app.config.from_pyfile(os.environ['YOURAPPLICATION_SETTINGS'])

        :param variable_name: name of the environment variable
        :param silent: set to `True` if you want silent failure for missing
                       files.
        :return: bool. `True` if able to load config, `False` otherwise.
        """
        rv = os.environ.get(variable_name)
        if not rv:
            if silent:
                return False
            raise RuntimeError('The environment variable %r is not set '
                               'and as such configuration could not be '
                               'loaded.  Set this variable and make it '
                               'point to a configuration file' %
                               variable_name)
        return self.from_pyfile(rv, silent=silent)

    def from_pyfile(self, filename, silent=False):
        """Updates the values in the config from a Python file.  This function
        behaves as if the file was imported as module with the
        :meth:`from_object` function.

        :param filename: the filename of the config.  This can either be an
                         absolute filename or a filename relative to the
                         root path.
        :param silent: set to `True` if you want silent failure for missing
                       files.

        .. versionadded:: 0.7
           `silent` parameter.
        """
        filename = os.path.join(self.root_path, filename)
        d = imp.new_module('config')
        d.__file__ = filename
        try:
            with open(filename) as config_file:
                exec(compile(config_file.read(), filename, 'exec'), d.__dict__)
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        self.from_object(d)
        return True

    def from_object(self, obj):
        """Updates the values from the given object.  An object can be of one
        of the following two types:

        -   a string: in this case the object with that name will be imported
        -   an actual object reference: that object is used directly

        Objects are usually either modules or classes.

        Just the uppercase variables in that object are stored in the config.
        Example usage::

            app.config.from_object('yourapplication.default_config')
            from yourapplication import default_config
            app.config.from_object(default_config)

        You should not use this function to load the actual configuration but
        rather configuration defaults.  The actual config should be loaded
        with :meth:`from_pyfile` and ideally from a location not within the
        package because the package might be installed system wide.

        :param obj: an import name or object
        """
        if isinstance(obj, string_types):
            obj = import_string(obj)
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, dict.__repr__(self))



# -*- coding: utf-8 -*-

from netkit.box import Box
from collections import OrderedDict

# 如果header字段变化，那么格式也会变化
HEADER_ATTRS = OrderedDict([
    ('magic', ('i', 2037952207)),
    ('version', ('h', 0)),
    ('packet_len', ('i', 0)),
    ('cmd', ('i', 0)),
    ('client_ip_num', ('4s', '')),
    ])


class Task(Box):
    header_attrs = HEADER_ATTRS
